/*
 * pmud - PMU daemon for Linux/PPC on Powerbooks.
 * At present this supports the PB2400/3400/3500,
 * the 1999 G3 Powerbooks (aka "Lombard"), and probably
 * the previous generation of G3 Powerbooks (aka "Wallstreet").
 *
 * Copyright 1999 Paul Mackerras.
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version
 *  2 of the License, or (at your option) any later version.
 */
#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/poll.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <linux/ioctl.h>
#include "pmu.h"

#define PMUD_VERSION		"0.2"

#define POWERLEVEL_SCRIPT	"/sbin/pwrctl"
#define POWER_FILE		"/etc/powerlevels"
#define PORT			879

int pmu_fd;
int adb_fd;

int pmu_version;
#define PMU_VERSION_3400	9
#define PMU_VERSION_WALLSTREET	10	/* a guess */
#define PMU_VERSION_LOMBARD	11

int powermode;			/* 0 for battery, 1 for AC */
int powerlevel = -1;		/* current power level setting */
int powerlevels[2] = { 1, 3 };	/* settings for battery & AC */
int chloralhydrate;		/* makes you sleepy */
int ntick;

/* Go to sleep if we have <= this many seconds of battery power remaining */
#define CRITICAL_TIMELEFT	300
/* Require power to be critically low for this many seconds before sleeping */
#define CRITICAL_TIME		30
int critical_time;		/* # seconds of critically low power */

char sock_welcome[64];

#define chk(x, mess)	if (x) { perror(mess); exit(1); }
#define novm(p, what)	if ((p) == 0) { \
			fprintf(stderr, "Couldn't allocate memory for "what); \
			exit(1); }

#define INBSZ	128
#define OUTBSZ	128

struct sockstate {
	int	fd;
	int	nstat;		/* number of statuses we can send */
	int	inb_cnt;
	char	inbuf[INBSZ];
	int	out_pending;
	char	*outptr;
	char	outbuf[OUTBSZ];
	struct sockstate *next;
	struct sockstate **prevp;
};

#define STAT_OUTSTANDING	2

struct sockstate *all_sockets;

struct timeval pmu_poll_time;

int pmu_op(unsigned char cmd, unsigned char *par, int npar,
	   void *res, int minres, const char *desc);
void get_pmu_version(void);
void poll_pmu(void);
void pmu_intr(int, void *, int);
void conn_sock(int, void *, int);
void read_sock(int, void *, int);
void delete_socket(struct sockstate *);
void do_cmd(struct sockstate *);
int send_socket(struct sockstate *, char *, int);
void send_more(int, void *, int);
int sock_write(struct sockstate *, char *, int);
void snooze(void);
void do_power(struct sockstate *, char *);
void set_power(int);
void run_program(char *, char *, char *);

struct pollfd *polls;
int npolls;
int nall_polls;
int ipoll_current;
struct fd_info {
	void	(*proc)(int, void *, int);
	void	*arg;
} *polli;

void poll_init(void)
{
	nall_polls = 10;
	polls = malloc(nall_polls * sizeof(struct pollfd));
	novm(polls, "poll table");
	polli = malloc(nall_polls * sizeof(struct fd_info));
	novm(polli, "poll info table");
	npolls = 0;
}

void add_fd(int fd, int events, void (*proc)(int, void *, int), void *arg)
{
	if (npolls >= nall_polls) {
		nall_polls = npolls + 10;
		polls = realloc(polls, nall_polls * sizeof(struct pollfd));
		polli = realloc(polli, nall_polls * sizeof(struct fd_info));
	}
	polls[npolls].fd = fd;
	polls[npolls].events = events;
	polls[npolls].revents = 0;
	polli[npolls].proc = proc;
	polli[npolls].arg = arg;
	++npolls;
}

void rem_fd(int fd, int events)
{
	int i;

	for (i = 0; i < npolls; ++i)
		if (polls[i].fd == fd && polls[i].events == events)
			break;
	if (i >= npolls)
		return;
	for (--npolls; i < npolls; ++i) {
		if (i == ipoll_current)
			--ipoll_current;
		polls[i] = polls[i+1];
		polli[i] = polli[i+1];
		++i;
	}
}

void do_poll(unsigned to)
{
	int i, nr;

	nr = poll(polls, npolls, to);
	if (nr < 0) {
		if (errno == EINTR)
			return;
		perror("poll");
		exit(1);
	}
	for (i = 0; i < npolls && nr > 0; ++i) {
		ipoll_current = i;
		if (polls[i].revents) {
			if (polls[i].revents & POLLNVAL)
				rem_fd(polls[i].fd, polls[i].events);
			else
				polli[i].proc(polls[i].fd, polli[i].arg,
					      polls[i].revents);
			--nr;
		}
		i = ipoll_current;
	}
}

main(int ac, char **av)
{
	struct timeval now, tleft;
	unsigned tms;
	int s, len, fl;
	struct sockaddr_in sin;
	FILE *f;

	poll_init();

	adb_fd = open("/dev/adb", O_RDWR);
	chk(adb_fd < 0, "Couldn't open /dev/adb");

	pmu_fd = open("/dev/pmu", O_RDWR);
	chk(pmu_fd < 0, "Couldn't open /dev/pmu");
	chk(fcntl(pmu_fd, F_GETFL, &fl) < 0, "fcntl(F_GETFL)");
	fl |= O_NONBLOCK;
	chk(fcntl(pmu_fd, F_SETFL, &fl) < 0, "fcntl(F_SETFL)");
	add_fd(pmu_fd, POLLIN, pmu_intr, NULL);

	s = socket(PF_INET, SOCK_STREAM, 0);
	chk(s < 0, "Couldn't create socket");

	sin.sin_family = PF_INET;
	sin.sin_port = htons(PORT);
	sin.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
	chk(bind(s, &sin, sizeof(sin)) < 0, "bind socket");
	chk(listen(s, 5) < 0, "socket listen");
	add_fd(s, POLLIN|POLLPRI, conn_sock, NULL);
	len = sizeof(sin);
	chk(getsockname(s, &sin, &len) < 0, "getsockname");

	get_pmu_version();
	sprintf(sock_welcome, "pmud %s %d\n", PMUD_VERSION, pmu_version);

	if (!(ac > 1 && strcmp(av[1], "debug"))) {
		int pid = fork();
		chk(pid < 0, "Couldn't fork");
		if (pid > 0)
			exit(0);
		chdir("/");
		setsid();
	}

	if ((f = fopen(POWER_FILE, "r")) != 0) {
		fscanf(f, "%d %d", &powerlevels[0], &powerlevels[1]);
		fclose(f);
	}

	gettimeofday(&pmu_poll_time, NULL);
	chloralhydrate = 0;
	for (;;) {
		gettimeofday(&now, NULL);
		if (!timercmp(&pmu_poll_time, &now, >))
			poll_pmu();
		if (chloralhydrate) {
			snooze();
			chloralhydrate = 0;
			critical_time = 0;
			continue;
		}
		gettimeofday(&now, NULL);
		timersub(&pmu_poll_time, &now, &tleft);
		tms = tleft.tv_sec * 1000 + tleft.tv_usec / 1000 + 1;
		do_poll(tms);
	}
}

int pmu_op(unsigned char cmd, unsigned char *par, int npar,
	   void *res, int minres, const char *desc)
{
	unsigned char data[32];
	int i, n;

	data[0] = 6;		/* packet for PMU */
	data[1] = cmd;
	for (i = 0; i < npar; ++i)
		data[i+2] = par[i];
	chk(write(adb_fd, data, npar+2) < 0, desc);
	n = read(adb_fd, data, sizeof(data));
	chk(n < 0, desc);
	if (n < minres + 1 || data[0] != 0) {
		fprintf(stderr, "huh? %d bytes from PMU:", n);
		for (i = 0; i < n; ++i)
			fprintf(stderr, " %.2x", data[i]);
		fprintf(stderr, "\n");
		return -1;
	}
	if (minres > 0)
		memcpy(res, data+1, minres);
	return n - 1;
}

void get_pmu_version(void)
{
	unsigned char vers;
	char env[16];

	if (pmu_op(PMU_GET_VERSION, 0, 0, &vers, 1, "version req") < 0)
		exit(1);
	pmu_version = vers;
	printf("PMU version %d: ", vers);
	switch (vers) {
	case PMU_VERSION_3400:
		printf("2400/3400/3500\n");
		break;
	case PMU_VERSION_WALLSTREET:
		printf("G3 (wallstreet)\n");
		break;
	case PMU_VERSION_LOMBARD:
		printf("1999 G3 (lombard)\n");
		break;
	default:
		if (vers > PMU_VERSION_LOMBARD) {
			printf("wow, later than lombard! :-)\n");
		} else {
			printf("something ancient, sorry. :-(\n");
			exit(1);
		}
	}
	/* set an environment variable for pwrctl script to use */
	sprintf(env, "PMUVERSION=%d", vers);
	putenv(env);
}

void poll_pmu(void)
{
	struct sockstate *ss, *ssnext;
	int n, pwr, vb, i;
	char status[64];
	unsigned char reply[16];
	unsigned char switches;
	struct timeval now;
	int timeleft = -1;

	if (pmu_op(PMU_GET_INPUT, 0, 0, &switches, 1, "switch req") < 0)
		return;

	if (pmu_version == PMU_VERSION_3400) {
		/* use extended battery request (6b) */
		unsigned char reply[8];

		if (pmu_op(PMU_EXT_BATT, 0, 0, reply, 8, "ext batt req") < 0)
			return;

		pwr = reply[0] & 1;			/* AC indicator */
		vb = (reply[1] << 8) + reply[2];	/* battery voltage */

		for (i = 0; i < 8; ++i)
			status[i] = ((reply[0] >> (7 - i)) & 1) + '0';
		sprintf(status+8, " %d %d %d %d %d %x\n", vb, reply[3],
			reply[4], reply[5], (reply[6] << 8) + reply[7],
			switches);

	} else {
		/* wallstreet/lombard, use smart battery request (6f) */
		signed short batt[5];
		unsigned char par;
		int i, j;
		int charge, current;
		char *p = status;

		*p++ = 'S';
		pwr = 0;
		charge = current = 0;
		for (i = 0; i < 2; ++i) {
			par = i + 1;
			if (pmu_op(PMU_SMART_BATT, &par, 1, batt, sizeof(batt),
				   "smart battery req") < 0)
				return;
			*p++ = ' ';
			*p++ = '{';
			for (j = 0; j < 3; ++j)
				*p++ = ((batt[0] >> j) & 1) + '0';
			pwr |= batt[0] & 1;
			if (batt[0] & 4) {
				/* have a battery */
				charge += batt[1];
				current += batt[3];
				for (j = 1; j < 5; ++j) {
					sprintf(p, " %d", batt[j]);
					p += strlen(p);
				}
			}
			*p++ = '}';
		}
		*p++ = '\n';
		*p = 0;
		if (current < 0)
			timeleft = charge * 3552 / -current;
	}

	n = strlen(status);
	for (ss = all_sockets; ss != 0; ss = ssnext) {
		ssnext = ss->next;
		if (ss->nstat >= STAT_OUTSTANDING)
			continue;
		switch (send_socket(ss, status, n)) {
		case 1:
			++ss->nstat;
			break;
		case -1:
			delete_socket(ss);
			break;
		}
	}

	if ((switches & 1) != 0)
		chloralhydrate = 1;	/* lid is closed, go to sleep */

	if (!pwr && timeleft >= 0 && timeleft < CRITICAL_TIMELEFT) {
		if (++critical_time >= CRITICAL_TIME) {
			fprintf(stderr, "battery critically low: sleeping\n");
			chloralhydrate = 1;
		}
	} else
		critical_time = 0;

	powermode = pwr;
	if (powerlevels[pwr] != powerlevel)
		set_power(powerlevels[pwr]);

	gettimeofday(&now, NULL);
	++pmu_poll_time.tv_sec;
	if (timercmp(&pmu_poll_time, &now, <)) {
		pmu_poll_time = now;
		++pmu_poll_time.tv_sec;
	}
}

void pmu_intr(int fd, void *arg, int ev)
{
	unsigned char intr[16];
	int i, n;
	char str[8];

	n = read(fd, intr, sizeof(intr));
	if (n <= 0) {
		if (n < 0)
			perror("read pmu");
		rem_fd(fd, POLLIN);
		return;
	}
	switch (intr[0]) {
	case 0x80:		/* tick interrupt */
		if (n != 1)
			break;
		++ntick;
		return;
	case 4:			/* PC card eject button */
		if (n < 3 || intr[2] < 1 || intr[2] > 2)
			break;
		sprintf(str, "%d", intr[2] - 1);
		run_program("/sbin/cardctl", "eject", str);
		return;
	}
	printf("PMU interrupt:");
	for (i = 0; i < n; ++i)
		printf(" %.2x", intr[i]);
	printf("\n");
}

int send_socket(struct sockstate *ss, char *p, int nb)
{
	int n;
	
	if (ss->out_pending != 0)
		return 0;
	n = sock_write(ss, p, nb);
	if (n < 0)
		return -1;
	if ((nb -= n) > 0) {
		memcpy(ss->outbuf, p + n, nb);
		ss->outptr = ss->outbuf;
		ss->out_pending = nb;
		add_fd(ss->fd, POLLOUT, send_more, ss);
	}
	return 1;
}

void send_more(int fd, void *arg, int ev)
{
	struct sockstate *ss = arg;
	int n;

	if (ss->out_pending == 0)
		return;
	n = sock_write(ss, ss->outptr, ss->out_pending);
	if (n < 0) {
		/* oops, looks bad */
		delete_socket(ss);
		return;
	}
	ss->outptr += n;
	ss->out_pending -= n;
	if (ss->out_pending == 0)
		rem_fd(fd, POLLOUT);
}

int sock_write(struct sockstate *ss, char *ptr, int nb)
{
	int n;

	n = write(ss->fd, ptr, nb);
	if (n < 0) {
		if (errno == EWOULDBLOCK || errno == EAGAIN)
			n = 0;
		else
			perror("write socket");
	}
	return n;
}

void conn_sock(int fd, void *arg, int ev)
{
	int fe, fl;
	struct sockstate *ss;

	fe = accept(fd, NULL, 0);
	chk(fcntl(fe, F_GETFL, &fl) < 0, "fcntl(F_GETFL)");
	fl |= O_NONBLOCK;
	chk(fcntl(fe, F_SETFL, &fl) < 0, "fcntl(F_SETFL)");
	ss = malloc(sizeof(*ss));
	if (ss == 0) {
		fprintf(stderr, "warning: no space for socket state\n");
		close(fe);
		return;
	}
	memset(ss, 0, sizeof(*ss));
	ss->fd = fe;
	ss->next = all_sockets;
	ss->prevp = &all_sockets;
	if (all_sockets != 0)
		all_sockets->prevp = &ss->next;
	all_sockets = ss;
	add_fd(fe, POLLIN, read_sock, ss);
	send_socket(ss, sock_welcome, strlen(sock_welcome));
}

void delete_socket(struct sockstate *ss)
{
	rem_fd(ss->fd, POLLIN);
	rem_fd(ss->fd, POLLOUT);
	close(ss->fd);
	*ss->prevp = ss->next;
	free(ss);
}

void read_sock(int fd, void *arg, int ev)
{
	struct sockstate *ss = arg;
	int n;
	int c = ss->inb_cnt;
	char *p, *endp;

	n = read(fd, ss->inbuf + c, sizeof(ss->inbuf) - c - 1);
	if (n <= 0) {
		if (n < 0) {
			if (errno == EWOULDBLOCK || errno == EAGAIN)
				return;
			if (errno != ECONNRESET)
				perror("read socket");
		}
		delete_socket(ss);
		return;
	}
	ss->inbuf[c + n] = 0;
	p = strchr(&ss->inbuf[c], '\n');
	c += n;
	if (p == 0) {
		if (c < sizeof(ss->inbuf) - 1) {
			ss->inb_cnt = c;
			return;
		}
	} else {
		c = p - ss->inbuf;
		*p++ = 0;
	}
	if (c > 0 && ss->inbuf[c-1] == '\r')
		ss->inbuf[c-1] = 0;
	do_cmd(ss);
	c = 0;
	if (p != 0 && *p != 0) {
		c = strlen(p);
		memmove(ss->inbuf, p, c);
	}
	ss->inb_cnt = c;
}

void do_cmd(struct sockstate *ss)
{
	char *p, *q;

	for (p = ss->inbuf; isspace(*p); ++p)
		;
	if (*p == 0) {
		/* acknowledgement of status */
		if (ss->nstat)
			--ss->nstat;
		return;
	}
	for (q = p; *q != 0 && !isspace(*q); ++q)
		;
	if (*q != 0) {
		for (*q++ = 0; isspace(*q); ++q)
			;
	}
	if (strcmp(p, "sleep") == 0)
		chloralhydrate = 1; 	/* go to sleep */
	else if (strcmp(p, "power") == 0)
		do_power(ss, q);
	/* else
	   printf("unknown cmd %s\n", p); */
}

void do_power(struct sockstate *ss, char *args)
{
	int p1, p2;
	char str[32];
	FILE *f;

	switch (sscanf(args, "%d %d", &p1, &p2)) {
	case -1:
	case 0:
		sprintf(str, "%d %d\n", powerlevels[0], powerlevels[1]);
		send_socket(ss, str, strlen(str));
		return;
	case 1:
		powerlevels[powermode] = p1;
		break;
	case 2:
		powerlevels[0] = p1;
		powerlevels[1] = p2;
		break;
	}

	f = fopen(POWER_FILE, "w");
	if (f != 0) {
		fprintf(f, "%d %d\n", powerlevels[0], powerlevels[1]);
		fclose(f);
	} else
		perror("Couldn't store power settings in " POWER_FILE);

	if (powerlevels[powermode] != powerlevel)
		set_power(powerlevels[powermode]);
}

void set_power(int level)
{
	char str[16];

	powerlevel = level;
	sprintf(str, "%d", level);
	run_program(POWERLEVEL_SCRIPT, str, 0);
}

void snooze(void)
{
	run_program(POWERLEVEL_SCRIPT, "s", 0);
	sync();
	if (ioctl(pmu_fd, PMU_IOC_SLEEP, 0) < 0)
		perror("PMU_IOC_SLEEP ioctl");
	gettimeofday(&pmu_poll_time, NULL);
	powerlevel = -1;
}

void run_program(char *prog, char *a1, char *a2)
{
	pid_t pid;
	int status = 0;

	pid = fork();
	if (pid == 0) {
		char *argv[4];

		argv[0] = prog;
		argv[1] = a1;
		argv[2] = a2;
		argv[3] = 0;
		execv(prog, argv);
		exit(99);
	}
	waitpid(pid, &status, 0);
	if (WIFEXITED(status)) {
		if (WEXITSTATUS(status) != 0)
			printf("%s exited with code %d\n", prog,
			       WEXITSTATUS(status));
	} else if (WIFSIGNALED(status)) {
		printf("%s got signal %d\n", prog, WTERMSIG(status));
	}
}
